;;; Termite code to allow slaves to connect and receive 
;;; instructions from the mast

(declare (block))

(define master-node '())

(define slave-pid-server '())

(define (start-master)
  (set! master-node (make-node (host-name) 7777)) 
  (node-init master-node)
  (init-slaves)
)

(define (init-slaves)
  (set! slave-pid-server 
	(spawn 
	 (lambda()
	   (with-exception-catcher
	    (lambda(e) (display (list 'error (object->string (handler e)))))
	    (lambda() (let* ((available-slaves (list master-node)))
			(let loop ((available-slaves available-slaves)
				   (used-slaves '())
				   (cloned-servers '()))
			  ; (termite-log-message "Slave PID server awaiting token")
			  (recv 
			   ((from tag ('assign-slave slave-host slave-port)) 
			    (! from (list tag 'assigned))
			    ; (termite-log-message ";; Assigned node " (list  slave-host slave-port))
			    (loop
			     (cons (make-node slave-host slave-port) available-slaves) used-slaves cloned-servers))

			   ((from tag ('next-slave)) 
			    (! from (list tag (car available-slaves)))
			    (if (null? (cdr available-slaves))
				(loop
				 (reverse (cons (car available-slaves) used-slaves)) '() cloned-servers)
				(loop 
				 (cdr available-slaves) (cons (car available-slaves) used-slaves) cloned-servers)))

			   ((from tag ('all-slaves)) 
			    (! from (list tag (append available-slaves used-slaves)))
			    (loop available-slaves used-slaves cloned-servers))
		
			   ;; Clone the slave server, and add it to the list of clones
			   ;; to whom new additions are to be added by receiving back the 
			   ;; clones pid. The receiving slave server shoould immediately reply
			   ;; with the add-clone message which will ripple out add-slave-clone
			   ;; messages to all other clones. So all clones and the master can use 
			   ;; added slave. It is possible to receive to clone messages before the 
			   ;; first clones add-clone messsage is recieved so the list of nodes 
			   ;; may not be exact but the fault should not be fatal.
			   ((from tag ('clone) )
			    ; (termite-log-message ";; Cloned slave server back to " from)
			    (call/cc 
			     (lambda (clone) 
			       (! from (list tag (lambda () (clone #t))))))
			    (loop available-slaves used-slaves cloned-servers))
		
			   (('add-clone cloned-slave slave-host slave-port)
			    ;; A new slave server has cloned the master. It is aware of the currently
			    ;; available nodes. We must now inform all other clones of its existence
			    (let ((cloned-node (make-node slave-host slave-port)))			      
			      ; (termite-log-message ";; Adding new slaver server " cloned-node)
			      ; (termite-log-message ";; Informing clone servers " cloned-servers)
			      (map (lambda(clone) (remote-spawn clone (lambda()
									(with-exception-catcher
									 (lambda(e) (display (list 'error (object->string (handler e)))))
									 (lambda() 
									   (add-slave-clone slave-host slave-port))))))
				   cloned-servers)
			      ; (termite-log-message ";; Added new slave server to all other servers " cloned-node)
			      ;; And only now add it the list of cloned-servers
			      (loop available-slaves used-slaves (cons cloned-node cloned-servers))))

			   (('add-slave-clone slave-host slave-port)
			    ;; A new slave server has cloned the master. It is aware of the currently
			    ;; available nodes. But we are not the master server so we need not inform
			    ;; any other slave. Instead we add it the list of available nodes.
			    ; (termite-log-message ";; Added node to server loop " (list slave-host slave-port))
			    (loop
			     (cons (make-node slave-host slave-port) available-slaves)
			     used-slaves
			     (cons (make-node slave-host slave-port) cloned-servers)))

			   (('remove-slave node)
			    (begin
			      ; (termite-log-message ";; Removed node " node)
			      (loop (remove node available-slaves) used-slaves cloned-servers))))))))))))

(define (close-slaves) (map (lambda(s) (remote-spawn s (lambda()(exit))))
			    all-slaves))

(define slave '())

(define (add-slave-clone slave-host slave-port)
  (termite-log-message ";; Adding slave clone " (list slave-host slave-port))
  (! slave-pid-server (list 'add-slave-clone slave-host slave-port)))

(define (assign-slave h p)
  (set! slave (!?  slave-pid-server (list 'assign-slave h p))))

(define (next-slave) 
  (!? slave-pid-server '(next-slave)))

(define (remove-slave node) 
  ;; In the event of an error the node believed to be the source of the error is 
  ;; removed from the slaves available to the host process an all others but
  ;; not at all if it is the master-node. This ensures that never mind where the
  ;; erorr occurs utlimately it will fall back to the master node and that node 
  ;; will always be available. 
  ;;
  ;; An attempt to remote-spawn to a failed node causes a bus error so we are best to
  ;; inform all nodes of the failure in the hope that none tries to remote spawn 
  ;; to the failed node. But this is not guaranteed.
  (newline)
  (map (lambda(s)
	 (display ";; Removing node ") (display node) (display " from ") (display s) (newline)
	 (remote-spawn s (lambda()
				   (if (not (eq? (get-master-node) node))
				       (! (get-slave-pid-server) (list 'remove-slave node))))))
       (remove node (all-slaves))))

;; Access master-node and slave-pid-server via functions
;; to ensure we pick up local copies of the global and not
;; copies migrated from the spawning process
(define (get-master-node) master-node)

(define (get-slave-pid-server) slave-pid-server)

(define (all-slaves) 
  (!? slave-pid-server '(all-slaves)))

(define (get-slave-count) 
  (if *distribute* 
      (length (all-slaves))
      1))

(define (clone-server to tag)
  (let ((clone (!?  slave-pid-server '(clone))))
    (! to (list tag clone))))

(define (server-is-cloned to-slave host port)
  (termite-log-message ";; Informed that server is cloned " (list to-slave host port))
  (with-exception-catcher
   (lambda(e) (display e))
   (lambda () 
     (! slave-pid-server (list 'add-clone to-slave host port))
     (termite-log-message ";; Server has been informed "))))
