;;; Utils to make use of Termites distributed processing capabilities

(declare (block))

;; To aid debugging disable spawning and run without error handler
(define *distribute* #t)

(define (d-on-off) (set! *distribute* (not *distribute*)))

(define (dpmap f lst nodf filter expect?)
  (let ((parent (self)))
       (map (lambda(tag-function x)
	      ;; To save network bandwith filter out values that do not require evaluation since they 
	      ;; evaluate to themselves
	      (let ((tag (nth 0 tag-function))
		    (node (nth 1 tag-function))
		    (function (nth 2 tag-function)))
		(cond
		 ((filter x) 
		  x)
		 ((not *distribute*) 
		  tag)
		 (else (recv 
			((,tag reply) 
			 (where (or (null? reply)
	                            (not (list? reply)) 
				    (not (every expect? reply))))
			 (newline)
			 (display ";; Unexpected reply to distributed call ") (write reply)
			 (newline)
			 (attempt-recovery node tag function ""))
			((,tag reply)
			 (termite-log-message "Received result for job tag" tag)
			 reply)
			;; If a slave fails to answer after 900 seconds assume it has failed
			;; so excute locally to recover the error, but beware this may cause
			;; errors if the 'failed' call has side effects
			(after 900 (attempt-recovery node tag function ";; Timeout on distributed map, giving up after waiting 900 seconds ")
))))))
	    (map (lambda (x) 
		   (cond
		    ((filter x) (list x '() '()))
		    ((not *distribute*) (list (f x) '() '()))
		    (else
		     (let* ((tag (make-tag))
			    (node (nodf))
			    (function-to-spawn (lambda()(! parent (list tag (f x))))))
		       (remote-spawn node function-to-spawn)
		       (termite-log-message "Spawned remote job" tag "to node" node)
		       (list tag node (lambda ()(f x)))))))
		 lst)
	    lst)))
		
(define (handler exc) 
  (cond ((type-exception? exc) 
	 (newline)(display ";;; Type exception ") 
	 (pp (list (type-exception-procedure exc) 
		   (type-exception-arguments exc) 
		   (type-exception-arg-num exc) 
		   (type-exception-type-id exc))))
	
	((unbound-global-exception? exc)
	 (newline)(display ";;; Unbound global ")
	 (display (unbound-global-exception-variable exc))
	 (newline))

	((error-exception? exc) 
	 (newline) (display (string-append ";; Error: " (error-exception-message exc)))
	 (display " ") (pretty-print (error-exception-parameters exc))
	 (newline))
	
	(else 
	 (newline)(display (string-append ";;; Unknown exception raised " (object->string exc)))))
  exc)

(define (attempt-recovery node tag function reason)
  (begin
    (display reason)
    (newline)
    (display ";; for node ")
    (display node)
    (display " to respond to tag ")
    (newline)
    (display ";; ")
    (display tag)
    (newline)
    ; (remove-slave node)
    (display ";; Evaluating locally in an attempt to recover")
    (newline)
    (let ((res '()))
      (dynamic-wind
	  (lambda() '())
	  (lambda()
	    (set! res (function))
	    res)
	  (lambda()
	    (begin
	      (display ";; Recovered for tag ")
	      (display tag)
	      (newline)
	      (display ";; With value ")
	      (write res)
	      (newline)))))))




(define (dpmap-split f lst nodf desc expect? . filter) 
  (let* ((filter (if (null? filter) filter (car filter)))
	 (nslaves (get-slave-count))
	 (jobs-per-slave (ceiling (/ (length lst) nslaves)))
	 (jobs (reduce append 
		       (if (null? filter)
			   (list
			      (partition lst jobs-per-slave))
			   (map (lambda(p)(filter-partion p filter))
				(partition lst jobs-per-slave))))))
    (let ((res (reduce append (map (lambda(r)(if (list? r) r (list r)))
				   (dpmap (lambda (l) (map f l)) 
					  jobs
					  nodf 
					  (lambda(x)(not (list? x)))
					  expect?)))))
      (if (or (null? res)
	      (not (list? res)) 
	      (not (every expect? res)))
	  (error "Unexpected type in parallel split" res))

      res)))


(define (partition lst psize)
  (cond
   ((null? lst)
    '())
   ((<= (length lst) psize)
    (list lst))
   (else
    (cons
     (butnthcdr psize lst)
     (partition (nthcdr psize lst) psize)))))
     

(define (filter-partion p f)
  ;; Break the partion upinto a list so that
  ;; elements for which f is false are in lists and others not
  ;; but the elements are still in the same order
  (cond
   ((null? p) '())
   ((f (car p)) (cons (car p) (filter-partion (cdr p) f)))
   (else
    (let ((remainder (filter-partion (cdr p) f)))
      (cond
       ((null? remainder) (list p))
       ((list? (car remainder)) (cons (cons (car p) (car remainder)) (cdr remainder)))
       (else (cons (list (car p)) remainder)))))))
